/* L# convex function -> submodular set function converter.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"
#include "sfm8.h"
#include "fw.h"

enum submodular_e submodular_type = IFF;

static int      is_natural;
static double   stored_func_val;
static double (*stored_func)(int,int*);
static int     *stored_x;
static int      stored_direction;
static int      stored_fix_index;

static int index(int *set, int i) {
    int r = set[i];
    if (r >= stored_fix_index) r++;
    return r;
}

static double oracle(int *set, int m, int dim) {
    int i;
    double f;

    if (m == 0) return 0;
    for (i=0; i<m; i++) stored_x[index(set,i)] += stored_direction;
    f = stored_func(dim+(!is_natural), stored_x);
    for (i=0; i<m; i++) stored_x[index(set,i)] -= stored_direction;
    oracle_call++;
#ifdef DEBUG2
    printf("%.14f (%g - %g) ", f-stored_func_val, f, stored_func_val);
    for (i=0; i<m; i++) printf("[%d]", set[i]);
    printf("\n");
#endif /* DEBUG2 */
    return f - stored_func_val;
}

/* OUTPUT: set[] */
double sf_minimize(int dim, double func(int,int*), int x[], int direction,
                   int set[], const int fix_index) {
    int i;
    double f1, f2;

    is_natural = (fix_index < 0);
    stored_func_val  = func(dim, x);
    stored_func      = func;
    stored_x         = x;
    stored_direction = direction;
    stored_fix_index = fix_index;
    if (is_natural) stored_fix_index = dim;
    else            dim--;

    switch (submodular_type) {
    case IFF: set[0] = iff_call(dim, oracle, set+1); break;
    case FW:  set[0] =  fw_call(dim, oracle, set+1); break;
    case IFF_FW:
        set[0] = iff_call(dim, oracle, set+1);
        f1 = oracle(set+1, set[0], dim);
        set[0] =  fw_call(dim, oracle, set+1);
        f2 = oracle(set+1, set[0], dim);
        if (fcmp(f1,f2) != 0) {
            printf("iff = %f, fw = %f\n", f1, f2);
            exit(1);
        }
        printf("ans = "); ivector_print(set[0], set+1); printf("\n");
        break;
    default:    assert(0!=0); break;
    }
    f1 = oracle(set+1, set[0], dim);
    if (!is_natural) {
        for (i=0; i<set[0]; i++) {
            if (set[i+1] >= fix_index) set[i+1]++;
        }
    }
    return f1;
}
